@@ -3,7 +3,7 @@
  *
  * pidgin-sipe
  *
- * Copyright (C) 2010-11 SIPE Project <http://sipe.sourceforge.net/>
+ * Copyright (C) 2010-2019 SIPE Project <http://sipe.sourceforge.net/>
  * Copyright (C) 2008 Novell, Inc.
  * Copyright (C) 2005 Thomas Butter <butter@uni-mannheim.de>
  *
@@ -32,6 +32,7 @@
 #include "sipmsg.h"
 #include "sipe-backend.h"
 #include "sipe-mime.h"
+#include "sipe-rtf.h"
 #include "sipe-utils.h"
 
 struct sipmsg *sipmsg_parse_msg(const gchar *msg) {
@@ -90,12 +91,25 @@ struct sipmsg *sipmsg_parse_header(const gchar *header) {
 		if (tmp && sipe_strcase_equal(tmp, "chunked")) {
 			msg->bodylen = SIPMSG_BODYLEN_CHUNKED;
 		} else {
-			SIPE_DEBUG_FATAL_NOFORMAT("sipmsg_parse_header(): Content-Length header not found");
+			tmp = sipmsg_find_content_type_header(msg);
+			if (tmp) {
+				/*
+				 * This is a fatal error situation: the message
+				 * is corrupted and we can't proceed. Set the
+				 * response code to a special value so that the
+				 * caller can abort correctly.
+				 */
+				SIPE_DEBUG_ERROR_NOFORMAT("sipmsg_parse_header: Content-Length header not found. Aborting!");
+				msg->response = SIPMSG_RESPONSE_FATAL_ERROR;
+				return(msg);
+			} else {
+				msg->bodylen = 0;
+			}
 		}
 	}
 	if(msg->response) {
 		const gchar *tmp;
-		tmp = sipmsg_find_header(msg, "CSeq");
+		tmp = sipmsg_find_cseq_header(msg);
 		if(!tmp) {
 			/* SHOULD NOT HAPPEN */
 			msg->method = 0;
@@ -170,24 +184,6 @@ char *sipmsg_to_string(const struct sipmsg *msg) {
 	return g_string_free(outstr, FALSE);
 }
 
-/**
- * Adds header to current message headers at specified position
- */
-void sipmsg_add_header_now_pos(struct sipmsg *msg, const gchar *name, const gchar *value, int pos) {
-	struct sipnameval *element = g_new0(struct sipnameval,1);
-
-	/* SANITY CHECK: the calling code must be fixed if this happens! */
-	if (!value) {
-		SIPE_DEBUG_ERROR("sipmsg_add_header_now_pos: NULL value for %s (%d)",
-				 name, pos);
-		value = "";
-	}
-
-	element->name = g_strdup(name);
-	element->value = g_strdup(value);
-	msg->headers = g_slist_insert(msg->headers, element,pos);
-}
-
 /**
  * Adds header to current message headers
  */
@@ -344,7 +340,7 @@ int sipmsg_parse_cseq(struct sipmsg *msg)
 {
 	int res = -1;
 	gchar **items;
-	items = g_strsplit(sipmsg_find_header(msg, "CSeq"), " ", 1);
+	items = g_strsplit(sipmsg_find_cseq_header(msg), " ", 1);
 	if (items[0]) {
 		res = atoi(items[0]);
 	}
@@ -401,7 +397,7 @@ void sipmsg_parse_p_asserted_identity(const gchar *header, gchar **sip_uri,
 	}
 
 	parts = g_strsplit(header, ",", 0);
-	
+
 	for (p = parts; *p; p++) {
 		gchar *uri = sipmsg_find_part_of_header(*p, "<", ">", NULL);
 		if (!uri)
@@ -442,7 +438,14 @@ void sipmsg_parse_p_asserted_identity(const gchar *header, gchar **sip_uri,
 const gchar *sipmsg_find_auth_header(struct sipmsg *msg, const gchar *name) {
 	GSList *tmp;
 	struct sipnameval *elem;
-	int name_len = strlen(name);
+	int name_len;
+
+	if (!name) {
+		SIPE_DEBUG_INFO_NOFORMAT("sipmsg_find_auth_header: no authentication scheme specified");
+		return NULL;
+	}
+
+	name_len = strlen(name);
 	tmp = msg->headers;
 	while(tmp) {
 		elem = tmp->data;
@@ -458,7 +461,7 @@ const gchar *sipmsg_find_auth_header(struct sipmsg *msg, const gchar *name) {
 		/* SIPE_DEBUG_INFO_NOFORMAT("moving to next header"); */
 		tmp = g_slist_next(tmp);
 	}
-	SIPE_DEBUG_INFO("auth header '%s' not found.", name);
+	SIPE_DEBUG_INFO("sipmsg_find_auth_header: '%s' not found", name);
 	return NULL;
 }
 
@@ -506,7 +509,7 @@ static gchar *sipmsg_get_x_mms_im_format(gchar *msgr) {
 	return x_mms_im_format;
 }
 
-gchar *sipmsg_get_msgr_string(gchar *x_mms_im_format) {
+gchar *sipmsg_get_msgr_string(const gchar *x_mms_im_format) {
 	gchar *msgr_orig;
 	gsize msgr_utf16_len;
 	gchar *msgr_utf16;
@@ -563,8 +566,9 @@ static void get_html_message_mime_cb(gpointer user_data,
 	if (!data->preferred) {
 		gboolean copy = FALSE;
 
-		/* preferred format */
-		if (g_str_has_prefix(type, "text/html")) {
+		/* preferred formats */
+		if (g_str_has_prefix(type, "text/html") ||
+		    g_str_has_prefix(type, "text/rtf")) {
 			copy = TRUE;
 			data->preferred = TRUE;
 
@@ -638,6 +642,10 @@ gchar *get_html_message(const gchar *ms_text_format_in, const gchar *body_in)
 				*d++ = c;
 		*d = c;
 
+	} else if (g_str_has_prefix(ms_text_format, "text/rtf")) {
+		char *tmp = res;
+		res = sipe_rtf_to_html(res);
+		g_free(tmp);
 	} else {
 		char *tmp = res;
 		res = g_markup_escape_text(res, -1); // as this is not html
@@ -713,6 +721,60 @@ sipmsg_parse_warning(struct sipmsg *msg, gchar **reason)
 	return code;
 }
 
+const gchar *sipmsg_find_call_id_header(const struct sipmsg *msg) {
+	return(sipmsg_find_header(msg, "Call-ID"));
+}
+
+const gchar *sipmsg_find_content_type_header(const struct sipmsg *msg) {
+	return(sipmsg_find_header(msg, "Content-Type"));
+}
+
+const gchar *sipmsg_find_cseq_header(const struct sipmsg *msg) {
+	return(sipmsg_find_header(msg, "CSeq"));
+}
+
+const gchar *sipmsg_find_event_header(const struct sipmsg *msg) {
+	return(sipmsg_find_header(msg, "Event"));
+}
+
+const gchar *sipmsg_find_expires_header(const struct sipmsg *msg) {
+	return(sipmsg_find_header(msg, "Expires"));
+}
+
+const gchar *sipmsg_find_from_header(const struct sipmsg *msg) {
+	return(sipmsg_find_header(msg, "From"));
+}
+
+const gchar *sipmsg_find_to_header(const struct sipmsg *msg) {
+	return(sipmsg_find_header(msg, "To"));
+}
+
+gchar *sipmsg_parse_address_from_header(const struct sipmsg *msg,
+					const gchar *name) {
+	return(parse_from(sipmsg_find_header(msg, name)));
+}
+
+gchar *sipmsg_parse_contact_address(const struct sipmsg *msg) {
+	return(sipmsg_parse_address_from_header(msg, "Contact"));
+}
+
+gchar *sipmsg_parse_from_address(const struct sipmsg *msg) {
+	return(sipmsg_parse_address_from_header(msg, "From"));
+}
+
+gchar *sipmsg_parse_to_address(const struct sipmsg *msg) {
+	return(sipmsg_parse_address_from_header(msg, "To"));
+}
+
+void sipmsg_update_to_header_tag(struct sipmsg *msg) {
+	const gchar *old = sipmsg_find_to_header(msg);
+	gchar *tag       = gentag();
+	gchar *new       = g_strdup_printf("%s;tag=%s", old, tag);
+	g_free(tag);
+	sipmsg_remove_header_now(msg, "To");
+	sipmsg_add_header_now(msg, "To", new);
+	g_free(new);
+}
 
 
 //------------------------------------------------------------------------------------------
@@ -886,6 +948,29 @@ sipe_parse_html(const char *html, char **attributes, char **message)
 	g_return_if_fail(attributes != NULL);
 	g_return_if_fail(message    != NULL);
 
+#define _HTML_UNESCAPE \
+	if (!g_ascii_strncasecmp(c, "&lt;", 4)) { \
+		msg[retcount++] = '<'; \
+		c += 4; \
+	} else if (!g_ascii_strncasecmp(c, "&gt;", 4)) { \
+		msg[retcount++] = '>'; \
+		c += 4; \
+	} else if (!g_ascii_strncasecmp(c, "&nbsp;", 6)) { \
+		msg[retcount++] = ' '; \
+		c += 6; \
+	} else if (!g_ascii_strncasecmp(c, "&quot;", 6)) { \
+		msg[retcount++] = '"'; \
+		c += 6; \
+	} else if (!g_ascii_strncasecmp(c, "&amp;", 5)) { \
+		msg[retcount++] = '&'; \
+		c += 5; \
+	} else if (!g_ascii_strncasecmp(c, "&apos;", 6)) { \
+		msg[retcount++] = '\''; \
+		c += 6; \
+	} else { \
+		msg[retcount++] = *c++; \
+	}
+
 	len = strlen(html);
 	msg = g_malloc0(len + 1);
 
@@ -903,6 +988,17 @@ sipe_parse_html(const char *html, char **attributes, char **message)
 				msg[retcount++] = '\n';
 				c += 4;
 			}
+			else if (!g_ascii_strncasecmp(c + 1, "div>", 4))
+			{
+				msg[retcount++] = '\r';
+				msg[retcount++] = '\n';
+				c += 5;
+				if (!g_ascii_strncasecmp(c, "<br></div>", 10)) {
+					/* This is an empty paragraph; replace it with
+					 * one line break. */
+					c += 10;
+				}
+			}
 			else if (!g_ascii_strncasecmp(c + 1, "i>", 2))
 			{
 				if (!has_italic)
@@ -947,7 +1043,10 @@ sipe_parse_html(const char *html, char **attributes, char **message)
 					c += 7;
 
 				while ((*c != '\0') && g_ascii_strncasecmp(c, "\">", 2))
-					msg[retcount++] = *c++;
+					if (*c == '&') {
+						_HTML_UNESCAPE;
+					} else
+						msg[retcount++] = *c++;
 
 				if (*c != '\0')
 					c += 2;
@@ -1060,38 +1159,7 @@ sipe_parse_html(const char *html, char **attributes, char **message)
 		}
 		else if (*c == '&')
 		{
-			if (!g_ascii_strncasecmp(c, "&lt;", 4))
-			{
-				msg[retcount++] = '<';
-				c += 4;
-			}
-			else if (!g_ascii_strncasecmp(c, "&gt;", 4))
-			{
-				msg[retcount++] = '>';
-				c += 4;
-			}
-			else if (!g_ascii_strncasecmp(c, "&nbsp;", 6))
-			{
-				msg[retcount++] = ' ';
-				c += 6;
-			}
-			else if (!g_ascii_strncasecmp(c, "&quot;", 6))
-			{
-				msg[retcount++] = '"';
-				c += 6;
-			}
-			else if (!g_ascii_strncasecmp(c, "&amp;", 5))
-			{
-				msg[retcount++] = '&';
-				c += 5;
-			}
-			else if (!g_ascii_strncasecmp(c, "&apos;", 6))
-			{
-				msg[retcount++] = '\'';
-				c += 6;
-			}
-			else
-				msg[retcount++] = *c++;
+			_HTML_UNESCAPE;
 		}
 		else
 			msg[retcount++] = *c++;
@@ -1106,6 +1174,8 @@ sipe_parse_html(const char *html, char **attributes, char **message)
 	*message = msg;
 
 	g_free(fontface);
+
+#undef _HTML_UNESCAPE
 }
 // End of TEMP
 
